# -- Project information -----------------------------------------------------

project = 'Django Courier'
copyright = '2018, Alan Trick'
author = 'Alan Trick'

# The short X.Y version
version = ''
# The full version, including alpha/beta/rc tags
release = '0.11.0'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.doctest',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

source_suffix = '.rst'

master_doc = 'index'

language = None

exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

pygments_style = 'sphinx'


# -- Options for HTML output -------------------------------------------------

html_theme = 'classic'
html_static_path = ['_static']
htmlhelp_basename = 'DjangoCourierdoc'

# -- Options for LaTeX output ------------------------------------------------

latex_elements = {}

latex_documents = [
    (master_doc, 'DjangoCourier.tex', 'Django Courier Documentation',
     'Alan Trick', 'manual'),
]


# -- Options for manual page output ------------------------------------------

man_pages = [
    (master_doc, 'djangocourier', 'Django Courier Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

texinfo_documents = [
    (master_doc, 'DjangoCourier', 'Django Courier Documentation',
     author, 'DjangoCourier', 'One line description of project.',
     'Miscellaneous'),
]



import collections
from typing import Any, List, Mapping, TypeVar, cast

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Q
from django.template import Context
from django.utils.translation import ugettext_lazy as _

from . import templates
from .backends import get_backends, get_backends_from_settings
from .base import AbstractContactable, Contact

_COURIER_CONTENTTYPE_IDS = None


def get_model_variables(label, value, cls, ancestors=set()):
    if cls is None:
        return {'label': label, 'value': value}
    assert issubclass(cls, models.Model)
    sub_ancestors = ancestors.copy()
    sub_ancestors.add(cls)
    attrs = []
    skip_relations = len(ancestors) > 2
    children = []
    for field in cls._meta.fields:
        sub_label = field.verbose_name.title(),
        sub_value = '{}.{}'.format(value, field.name)
        if field.is_relation:
            if skip_relations or field.related_model in ancestors:
                continue
            children.append(get_model_variables(
                sub_label, sub_value, field.related_model,
                ancestors=sub_ancestors))
        else:
            attrs.append({'label': sub_label, 'value': sub_value})
    return {'label': label, 'value': value, 'attrs': attrs,
            'rels': children}


def _courier_contenttype_ids():
    for all_models in apps.all_models.values():
        for model in all_models.values():
            if issubclass(model, CourierModel):
                if model._courier_meta.has_channels():
                    ct = ContentType.objects.get_for_model(model)
                    yield ct.id


def content_type_limit():
    global _COURIER_CONTENTTYPE_IDS
    if _COURIER_CONTENTTYPE_IDS is None:
        _COURIER_CONTENTTYPE_IDS = tuple(_courier_contenttype_ids())
    return {'id__in': _COURIER_CONTENTTYPE_IDS}


class Channel(collections.namedtuple('ChannelBase',
                                     ('name', 'cls', 'func', 'obj'))):
    def contactables(self):
        return (self.obj,) if self.func is None else self.func(self.obj)


class CourierOptions(object):
    """
    Options for Courier extensions
    """

    PREFIX_NAMES = {
        'si': _('Site Contacts'),
        'c': 'Content',
        'se': _('Source'),
        're': _('Target'),
    }
    PREFIX_FORMATS = {
        'c': '{}',
        'se': _('Source\'s {}'),
        're': _('Target\'s {}'),
    }
    ALL_OPTIONS = ('notifications', 'channels')
    # list of notification code names
    notifications = []

    def __init__(self, meta):
        """
        Set any options provided, replacing the default values
        """
        self.__channel_items = None
        self._channels = {}
        self.notifications = []
        if meta is not None:
            for key, value in meta.__dict__.items():
                if key in self.ALL_OPTIONS:
                    setattr(self, key, value)
                elif not key.startswith('_'):  # ignore private parts
                    raise ValueError(
                        'CourierMeta has invalid attribute: {}'.format(key))

    def get_channels(self, prefix: str, obj) -> dict:
        if self.__channel_items is None:
            self.__channel_items = {}
            for key, (name, cls, func) in self._channels.items():
                channel_key = prefix if key == '' else prefix + ':' + key
                if name == '':
                    name = self.PREFIX_NAMES[prefix]
                    if name == 'Content':
                        name = cls._meta.verbose_name.title()
                else:
                    name = self.PREFIX_FORMATS[prefix].format(name)
                self.__channel_items[channel_key] = (name, cls, func)
        return dict((key, Channel(name, cls, func, obj))
                    for key, (name, cls, func) in self.__channel_items.items())

    def add_channel(self, key, name, target_type, func=None):
        self.__channel_items = None
        self._channels[key] = name, target_type, func

    def has_channels(self):
        return bool(self._channels)


class CourierModelBase(models.base.ModelBase):
    """
    Metaclass for Courier extensions.

    Deals with notifications on CourierOptions
    """
    def __new__(mcs, name, bases, attributes):
        new = super(CourierModelBase, mcs).__new__(
            mcs, name, bases, attributes)
        meta = attributes.pop('CourierMeta', None)
        setattr(new, '_courier_meta', CourierOptions(meta))
        return new


CourierModelN = TypeVar('CourierModelN', 'CourierModel', None)


class CourierModel(models.Model, metaclass=CourierModelBase):
    """
    Base class for models that implement notifications
    """

    class Meta:
        abstract = True

    def issue_notification(self, codename: str,
                           target: CourierModelN = None,
                           source: CourierModelN = None):
        ct = ContentType.objects.get_for_model(self)
        notification = Notification.objects.get(
            codename=codename, content_type=ct)
        notification.issue(self, target, source)

    @classmethod
    def add_channel(cls, key, verbose_name='', target_type=None, func=None):
        if target_type is None:
            target_type = cls
        elif func is None and target_type != cls:
            raise ValueError('Must specify function when using '
                             'different models')
        cls._courier_meta.add_channel(key, verbose_name, target_type, func)

    def get_channels(self, prefix: str):
        return self.__class__._courier_meta.get_channels(prefix, self)


class NotificationManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, app_label, model, codename):
        return self.get(
            codename=codename,
            content_type=ContentType.objects.db_manager(
                self.db).get_by_natural_key(app_label, model),
        )


class CourierParam:
    REQUIRED_PARAMS = {'codename', 'description'}
    OPTIONAL_PARAMS = {'source_model': '', 'target_model': '',
                       'required': False}

    def __init__(self, codename, description, **kwargs):
        self.params = {
            'codename': codename,
            'description': description,
            'from_code': True,
        }
        for key, default in self.OPTIONAL_PARAMS.items():
            if key in kwargs:
                self.params[key] = kwargs.pop(key)
            else:
                self.params[key] = default
        if kwargs:
            raise ValueError('Unrecognized parameters {}'.format(
                ', '.join(kwargs.keys())))

    def params_equal(self, notification):
        for key in self.params:
            if getattr(notification, key) != self.params[key]:
                return False
        return True

    def param_value(self, key):
        value = self.params[key]
        if key in ('source_model', 'target_model'):
            if value == '':
                return None
            model = apps.get_model(value)
            return ContentType.objects.get_for_model(model)
        return value

    def set_params(self, notification):
        for key in self.params:
            setattr(notification, key, self.param_value(key))

    def create(self, content_type):
        new = Notification(content_type=content_type)
        self.set_params(new)
        return new

    @property
    def codename(self):
        return self.params['codename']


class Notification(models.Model):
    """
    Base class for all notifications
    """

    codename = models.CharField(_('codename'), max_length=100)
    content_type = models.ForeignKey(
        to=ContentType, on_delete=models.CASCADE,
        limit_choices_to=content_type_limit,
        verbose_name=_('content type'))
    description = models.TextField(_('description'))
    source_model = models.ForeignKey(
        to=ContentType, on_delete=models.CASCADE, related_name='+',
        limit_choices_to=content_type_limit,
        verbose_name=_('source model'), null=True, blank=True)
    target_model = models.ForeignKey(
        to=ContentType, on_delete=models.CASCADE, related_name='+',
        limit_choices_to=content_type_limit,
        verbose_name=_('target model'), null=True, blank=True)
    required = models.BooleanField(
        _('required'), default=False,
        help_text=_('If true, triggering the notification will throw an '
                    'error if there is no available template/contact'))
    from_code = models.BooleanField(
        _('from code'), default=False,
        help_text=_('True if the notification is defined in the code and '
                    'automatically managed'))

    objects = NotificationManager()

    def __str__(self):
        return "{} | {} | {}".format(
            self.content_type.app_label, self.content_type, self.codename)

    def natural_key(self):
        return self.content_type.natural_key() + (self.codename,)

    natural_key.dependencies = ['contenttypes.contenttype']

    def get_recipient_models(self):
        source_model = self.get_source_model()
        target_model = self.get_target_model()
        content_model = (None if self.content_type_id is None
                         else self.content_type.model_class())
        recipient_spec = {
            'si': SiteContact,
            're': target_model,
            'se': source_model,
            'c': content_model,
        }
        return dict((key, value) for (key, value)
                    in recipient_spec.items() if value is not None)

    def get_recipient_instances(self, content, target, source):
        choices = {
            'si': SiteContact(),
            'c': content,
        }
        if self.target_model and (target is not None):
            choices['re'] = target
        elif self.target_model:
            raise RuntimeError(
                'Model specified "target_model" for notification but '
                'target missing on issue_notification ')
        elif target is not None:
            raise RuntimeError('Recipient added to issue_notification, '
                               'but is not specified in CourierMeta')
        if self.source_model and (source is not None):
            choices['se'] + source
        elif self.source_model:
            raise RuntimeError(
                'Model specified "source_model" for notification but source '
                'missing on issue_notification ')
        elif source is not None:
            raise RuntimeError('Sender added to issue_notification, but is '
                               'not specified in CourierMeta')

        return dict((key, model) for (key, model)
                    in choices.items() if model is not None)

    def get_recipient_channels(self, content, target, source):
        instances = self.get_recipient_instances(content, target, source)
        return dict((key, channel)
                    for recip_key, model in instances.items()
                    for key, channel in model.get_channels(recip_key).items())

    def issue(self, content,
              target: CourierModelN=None,
              source: CourierModelN=None):
        """
        To send a notification to a user, get all the user's active methods.
        Then get the backend for each method and find the relevant template
        to send (and has the said notification). Send that template with
        the parameters with the backend.

        :param content: model object that the notification is about
        :param target: either a user, or None if no logical target
        :param source: user who initiated the notification
        :return: None
        """
        # check
        parameters = {'content': content}
        if target is not None:
            parameters['target'] = target
        if source is not None:
            parameters['source'] = source

        channels = self.get_recipient_channels(content, target, source)
        contactable_list = dict((key, channel.contactables())
                                for (key, channel) in channels.items())

        contact_set = collections.defaultdict(dict)
        for key, contactables in contactable_list.items():
            for c_able in contactables:
                for contact in c_able.get_contacts_for_notification(self):
                    contact_set[key][contact] = c_able

        sent = False
        for key, contact_dict in contact_set.items():
            if self.send_messages(contact_dict, parameters,
                                  Template.objects.filter(recipient=key)):
                sent = True
        if not sent and self.required:
            raise RuntimeError('Notification required, but no message sent')

    def send_messages(
            self, contacts: Mapping[Contact, AbstractContactable],
            generic_params: Mapping[str, Any], template_queryset):

        def _get_backend_message(protocol):
            backends = get_backends_from_settings(protocol)
            # now get all the templates for these backends
            for be in backends:
                tpl = template_queryset.filter(
                    backend=be.ID, notification=self, is_active=True).first()
                if tpl is not None:
                    return be, tpl
            return None

        # per-protocol message cache
        cache = {}
        sent = False
        for contact, contactable in contacts.items():
            params = generic_params.copy()
            params['contact'] = contact
            params['recipient'] = contactable
            proto = contact.protocol
            if proto not in cache:
                cache[proto] = _get_backend_message(proto)
            if cache[proto] is not None:
                backend, template = cache[proto]
                message = backend.build_message(
                    template.subject, template.content, params)
                # We're catching all exceptions here because some people
                # are bad people and can't subclass properly
                try:
                    backend.send_message(contact, message)
                    sent = True
                except Exception as e:
                    FailedMessage.objects.create(
                        backend=backend.ID,
                        name=contact.name,
                        address=contact.address,
                        message=str(message),
                        error=str(e),
                    )
        return sent

    def can_issue_custom(self):
        return not (self.source_model or self.target_model)

    def preview(self, backend_id, message):
        backend = [b for b in get_backends() if b.ID == backend_id][0]
        content_model = self.content_type.model_class()
        params = templates.PreviewParameters(
            content_model, self.get_source_model(), self.get_target_model())
        return backend.preview_message('', message, params)

    def get_source_model(self):
        if self.source_model:
            return self.source_model.model_class()
        return None

    def get_target_model(self):
        if self.target_model:
            return apps.get_model(self.target_model)
        return None

    def get_recipient_variables(self):
        recipient_spec = self.get_recipient_models()
        source_model = self.get_source_model()
        target_model = self.get_target_model()
        content_model = self.content_type.model_class()
        mapping = {}
        for target_key, model in recipient_spec.items():
            if model is not None:
                channel_items = model._courier_meta.get_channels(
                    target_key, None).items()
                for key, (name, cls, _func, _obj) in channel_items:
                    mapping[key] = get_model_variables(
                        'Recipient', 'recipient', cls)
        content_name = content_model._meta.verbose_name.title()
        mapping['_static'] = [
            get_model_variables(content_name, 'content', content_model),
        ]
        if source_model:
            mapping['_static'].append(
                get_model_variables('Source', 'source', source_model))
        if target_model:
            mapping['_static'].append(
                get_model_variables('Target', 'target', target_model))
        return mapping


class Template(models.Model):

    class Meta:
        verbose_name = _('template')

    notification = models.ForeignKey(
        to=Notification, on_delete=models.PROTECT,
        verbose_name=_('notification'))
    backend = models.CharField(_('backend'), max_length=100)
    subject = models.CharField(_('subject'), max_length=500, blank=True)
    content = models.TextField(_('content'))
    recipient = models.CharField(
        verbose_name=_('recipient'), max_length=103, default='re',
        help_text=_('Who this message actually gets sent to.'))
    is_active = models.BooleanField(
        _('is active'), default=True,
        help_text=_('When not active, the template will be ignored'))

    objects = models.Manager()

    def render(self, parameters: dict, autoescape=True):
        content = cast(str, self.content)
        template = templates.from_string(content)
        context = Context(parameters, autoescape=autoescape)
        return template.render(context)


class SiteContactManager(models.Manager, AbstractContactable):
    use_in_migrations = True

    def get_contacts_for_notification(
            self, notification: 'Notification') -> List[Contact]:
        wlq = Q(enable_filter='whitelist',
                sitecontactpreference__notification=notification,
                sitecontactpreference__is_active=False)
        blq = Q(enable_filter='blacklist') & ~Q(
                sitecontactpreference__notification=notification,
                sitecontactpreference__is_active=False)
        for sc in SiteContact.objects.filter(blq | wlq):
            yield Contact(sc.name, sc.protocol, sc.address)


# can't make this subclass AbstractContact or fields become unset-able
class SiteContact(CourierModel):

    ENABLE_CHOICES = (
        ('blacklist', _('Blacklist')),
        ('whitelist', _('Whitelist')),
    )

    class Meta:
        verbose_name = _('site contact')
        unique_together = (('address', 'protocol'),)

    name = models.CharField(_('name'), blank=True, max_length=500)
    protocol = models.CharField(_('protocol'), max_length=100)
    address = models.CharField(_('address'), max_length=500)
    enable_filter = models.CharField(choices=ENABLE_CHOICES, max_length=10,
                                     default='blacklist')

    objects = SiteContactManager()

    def __str__(self):
        return self.name

    @staticmethod
    def all_contacts(_obj):
        yield SiteContact.objects


SiteContact.add_channel('', func=SiteContact.all_contacts)


class SiteContactPreference(models.Model):

    site_contact = models.ForeignKey(SiteContact, on_delete=models.CASCADE)
    notification = models.ForeignKey(Notification, on_delete=models.CASCADE)
    is_active = models.BooleanField(_('is active'))

    objects = models.Manager()


class FailedMessage(models.Model):

    class Meta:
        verbose_name = _('failed message')

    backend = models.CharField(_('backend'), max_length=100)
    contact_name = models.CharField(_('contact name'), max_length=500)
    address = models.CharField(_('address'), max_length=500)
    message = models.TextField(_('message'))
    error = models.TextField(_('error'))
    created_at = models.DateTimeField(_('created at'), auto_now_add=True)

    objects = models.Manager()

    def __str__(self):
        return '{}:{} @ {}'.format(self.backend.PROTOCOL,
                                   self.address, self.created_at)


def get_recipient_choices(notification):
    for recipient_key, model in notification.get_recipient_models().items():
        channel_data = model._courier_meta.get_channels(recipient_key, None)
        for key, (name, _cls, _func, _obj) in channel_data.items():
            yield key, name



import collections
import json
import pydoc

import django.conf
import django.core.mail.backends.base
import django.core.mail.backends.smtp
import django.template
import django.utils.html
import requests
from django.template import Context
from django.utils.translation import ugettext_lazy as _

from . import settings, templates

PROTOCOLS = {
    'email': _('Email'),
    'sms': _('SMS'),
    'slack-webhook': _('Slack Webhook'),
}

# a global that we initialize once, please use
# `get_backends_from_settings(str)`
__ALL_BACKENDS__ = collections.defaultdict(list)


def html_format(text: str):
    escaped = django.utils.html.escape(text)
    return escaped.replace('\r\n', '<br/>').replace('\n', '<br/>')


class NotificationBackend:

    USE_SUBJECT = False
    ESCAPE_HTML = True

    @classmethod
    def build_message(cls, subject: str, body: str, parameters: dict):
        template = templates.from_string(body)
        context = Context(parameters, autoescape=cls.ESCAPE_HTML)
        return template.render(context)

    @classmethod
    def preview_message(cls, subject: str, body: str, parameters: dict):
        message = cls.build_message(subject, body, parameters)
        if not cls.ESCAPE_HTML:
            message = html_format(message)
        return message

    @classmethod
    def send_message(
            cls, contact: 'django_courier.base.Contact', message):
        raise NotImplementedError


class BasicEmailBackend(NotificationBackend):

    ID = 'email-basic'
    PROTOCOL = 'email'
    VERBOSE_NAME = _('Email (Basic)')
    USE_SUBJECT = True

    @classmethod
    def send_message(cls, contact, message):
        mpm = templates.MultipartMessage.from_string(message)
        email = mpm.to_mail()
        email.from_email = django.conf.settings.DEFAULT_FROM_EMAIL
        email.to = [contact.address]
        connection = django.core.mail.get_connection()
        connection.send_messages([email])


class EmailBackend(BasicEmailBackend):

    ID = 'email'
    VERBOSE_NAME = _('Email (Template-based)')

    @classmethod
    def build_message(cls, subject: str, body: str, parameters: dict):
        return str(templates.email_parts(subject, body, parameters))

    @classmethod
    def preview_message(cls, subject: str, body: str, parameters: dict):
        parts = templates.email_parts(subject, body, parameters)
        if parts.html:
            return parts.html
        return django.utils.html.escape(parts.text)


class HtmlEmailBackend(BasicEmailBackend):

    ID = 'email-html'
    VERBOSE_NAME = _('Email (HTML)')

    @classmethod
    def build_message(cls, subject: str, body: str, parameters: dict):
        return str(templates.email_html(subject, body, parameters))

    @classmethod
    def preview_message(cls, subject: str, body: str, parameters: dict):
        return templates.email_html(subject, body, parameters).html


class MarkdownEmailBackend(BasicEmailBackend):

    ID = 'email-md'
    VERBOSE_NAME = _('Email (Markdown)')

    @classmethod
    def build_message(cls, subject: str, body: str, parameters: dict):
        return str(templates.email_md(subject, body, parameters))

    @classmethod
    def preview_message(cls, subject: str, body: str, parameters: dict):
        return templates.email_md(subject, body, parameters).html


class PostmarkTemplateBackend(NotificationBackend):

    ID = 'postmark_template'
    PROTOCOL = 'email'
    VERBOSE_NAME = _('Postmark Email')

    @classmethod
    def build_message(cls, subject: str, body: str, parameters: dict):
        return json.dumps({
            'template_id': body.content.strip(),
            'parameters': dict((k, str(v)) for k, v in parameters.items()),
        })

    @classmethod
    def send_message(cls, contact, message):
        from anymail.message import AnymailMessage
        data = json.loads(message)
        from_email = django.conf.settings.DEFAULT_FROM_EMAIL
        email = AnymailMessage('', '', from_email, [contact.address])
        email.template_id = data['template_id']
        email.merge_global_data = data['parameters']
        # TODO: implement
        # backend.send_messages([email])


class TwilioBackend(NotificationBackend):

    ID = 'twilio'
    PROTOCOL = 'sms'
    ESCAPE_HTML = False
    VERBOSE_NAME = _('Twilio')

    @classmethod
    def send_message(cls, contact, message):
        from twilio.rest import Client
        if not hasattr(django.conf.settings, 'TWILIO_ACCOUNT_SID'):
            raise django.conf.ImproperlyConfigured(
                'Twilio backend enabled but TWILIO_* settings missing')
        account_sid = django.conf.settings.TWILIO_ACCOUNT_SID
        auth_token = django.conf.settings.TWILIO_AUTH_TOKEN
        from_number = django.conf.settings.TWILIO_FROM_NUMBER
        client = Client(account_sid, auth_token)
        client.messages.create(
            to=contact.address, from_=from_number, body=message)


class SlackWebhookBackend(NotificationBackend):

    ID = 'slack-webhook'
    PROTOCOL = 'slack-webhook'
    ESCAPE_HTML = False
    VERBOSE_NAME = _('Slack')

    @classmethod
    def send_message(cls, contact, message):
        data = json.dumps({'text': message})
        headers = {'Content-Type': 'application/json'}
        result = requests.post(contact.address, data=data, headers=headers)
        if not result.ok:
            raise requests.HTTPError(result.text)


def _init_backends():
    """
    Intialize backend settings
    """
    for name in settings.BACKENDS:
        cls = pydoc.locate(name)
        __ALL_BACKENDS__[cls.PROTOCOL].append(cls)


def get_backends_from_settings(protocol: str):
    if not __ALL_BACKENDS__:
        _init_backends()

    for backend in __ALL_BACKENDS__[protocol]:
        yield backend


def get_backends():
    if not __ALL_BACKENDS__:
        _init_backends()
    for backend_list in __ALL_BACKENDS__.values():
        for backend in backend_list:
            yield backend


def get_backend_choices():
    if not __ALL_BACKENDS__:
        _init_backends()
    backends = []
    for backend_list in __ALL_BACKENDS__.values():
        for backend in backend_list:
            backends.append(backend)

    return ((bk.ID, bk.VERBOSE_NAME) for bk in backends)


def get_protocol_choices():
    if not __ALL_BACKENDS__:
        _init_backends()
    for protocol in __ALL_BACKENDS__.keys():
        yield (protocol, PROTOCOLS.get(protocol, protocol))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2018-04-08 05:07
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_courier', '0008_notification_required'),
    ]

    operations = [
        migrations.AddField(
            model_name='template',
            name='subject',
            field=models.CharField(
                blank=True, max_length=500, verbose_name='subject'),
        ),
        migrations.AlterField(
            model_name='template',
            name='backend',
            field=models.CharField(max_length=100, verbose_name='backend'),
        ),
        migrations.AlterField(
            model_name='template',
            name='content',
            field=models.TextField(verbose_name='content'),
        ),
        migrations.AlterField(
            model_name='template',
            name='is_active',
            field=models.BooleanField(
                default=True,
                help_text='When not active, the template will be ignored',
                verbose_name='is active'),
        ),
        migrations.AlterField(
            model_name='template',
            name='target',
            field=models.CharField(
                default='re',
                help_text='Who this message actually gets sent to.',
                max_length=103, verbose_name='target'),
        ),
    ]



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_courier', '0007_migrations'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='required',
            field=models.BooleanField(
                default=False,
                help_text='If true, triggering the notification will throw an'
                          ' error if there is no available template/contact',
                verbose_name='required'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


def forwards_func(apps, schema_editor):
    # We get the model from the versioned app registry;
    # if we directly import it, it'll be the wrong version
    template_cls = apps.get_model("django_courier", "Template")
    db_alias = schema_editor.connection.alias
    template_cls.objects.using(db_alias).filter(
        send_to_site_contacts=True).update(target='si')


def reverse_func(apps, schema_editor):
    # forwards_func() creates two Country instances,
    # so reverse_func() should delete them.
    template_cls = apps.get_model("django_courier", "Template")
    db_alias = schema_editor.connection.alias
    template_cls.objects.using(db_alias).filter(
        target='si').update(send_to_site_contacts=True)


class Migration(migrations.Migration):

    dependencies = [
        ('django_courier', '0003_failed_messages'),
    ]

    operations = [
        # uncomment to run on postgres
        # migrations.RunSQL('SET CONSTRAINTS ALL IMMEDIATE',
        #                   reverse_sql=migrations.RunSQL.noop),
        migrations.AddField(
            model_name='template',
            name='target',
            field=models.CharField(
                choices=[('re', 'Recipient'), ('si', 'Site Contacts'),
                         ('se', 'Sender')],
                default='re',
                help_text='Who this message actually gets sent to.',
                max_length=2),
        ),
        migrations.RunPython(forwards_func, reverse_func),
        # migrate target data
        migrations.RemoveField(
            model_name='template',
            name='send_to_site_contacts',
        ),
        migrations.AlterField(
            model_name='failedmessage',
            name='address',
            field=models.CharField(max_length=500, verbose_name='address'),
        ),
        migrations.AlterField(
            model_name='failedmessage',
            name='backend',
            field=models.CharField(max_length=100, verbose_name='backend'),
        ),
        migrations.AlterField(
            model_name='failedmessage',
            name='created_at',
            field=models.DateTimeField(
                auto_now_add=True, verbose_name='created at'),
        ),
        migrations.AlterField(
            model_name='failedmessage',
            name='error',
            field=models.TextField(verbose_name='error'),
        ),
        migrations.AlterField(
            model_name='failedmessage',
            name='message',
            field=models.TextField(verbose_name='message'),
        ),
        migrations.AlterField(
            model_name='failedmessage',
            name='protocol',
            field=models.CharField(max_length=100, verbose_name='protocol'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='description',
            field=models.TextField(verbose_name='description'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='use_recipient',
            field=models.BooleanField(verbose_name='use recipient'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='use_sender',
            field=models.BooleanField(verbose_name='use sender'),
        ),
        migrations.AlterField(
            model_name='sitecontact',
            name='address',
            field=models.CharField(max_length=500, verbose_name='address'),
        ),
        migrations.AlterField(
            model_name='sitecontact',
            name='protocol',
            field=models.CharField(max_length=100, verbose_name='protocol'),
        ),
        migrations.AlterField(
            model_name='sitecontactpreference',
            name='is_active',
            field=models.BooleanField(verbose_name='is active'),
        ),
        # uncomment to run on postgres
        # migrations.RunSQL('SET CONSTRAINTS ALL IMMEDIATE',
        #                   reverse_sql=migrations.RunSQL.noop),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-02-01 06:31
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_courier', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='SiteContact',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True,
                    serialize=False, verbose_name='ID')),
                ('address', models.CharField(max_length=500)),
                ('protocol', models.CharField(max_length=100)),
            ],
            options={
                'verbose_name': 'site contact',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='SiteContactPreference',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True,
                    serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField()),
                ('notification', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    to='django_courier.Notification')),
                ('site_contact', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    to='django_courier.SiteContact')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='contact',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='contact',
            name='user',
        ),
        migrations.AddField(
            model_name='template',
            name='send_to_site_contacts',
            field=models.BooleanField(
                default=False,
                help_text='Whether this message is sent to the site contacts '
                          'or to the notification recipient.'),
        ),
        migrations.DeleteModel(
            name='Contact',
        ),
        migrations.AlterUniqueTogether(
            name='sitecontact',
            unique_together={('address', 'protocol')},
        ),
    ]



import django.contrib.messages
import django.forms.utils
import django.http
from django import forms
from django.conf.urls import url
from django.contrib import admin
from django.contrib.admin.options import IS_POPUP_VAR
from django.contrib.admin.utils import unquote
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.core.exceptions import PermissionDenied, ValidationError
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.html import escape
from django.utils.translation import ugettext
from django.utils.translation import ugettext_lazy as _

from . import backends, models


class RecipientFilter(admin.SimpleListFilter):
    """A really simple filter so that we have the right labels
    """
    title = _('recipient')
    parameter_name = 'recipient'

    def lookups(self, request, model_admin):
        """Return list of key/name tuples
        """
        return models.get_recipient_choices()

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(recipient=self.value())
        return queryset


class SelectWithSubjectData(forms.Select):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.use_subjects = {}

    def create_option(self, name, value, label, selected,
                      index, subindex=None, attrs=None):
        index = str(index) if subindex is None else "%s_%s" % (index, subindex)
        option_attrs = {'data-subject': ('true' if self.use_subjects.get(value)
                                         else 'false')}
        if selected:
            option_attrs.update(self.checked_attribute)
        if 'id' in option_attrs:
            option_attrs['id'] = self.id_for_label(option_attrs['id'], index)
        return {
            'name': name,
            'value': value,
            'label': label,
            'selected': selected,
            'index': index,
            'attrs': option_attrs,
            'type': self.input_type,
            'template_name': self.option_template_name,
        }


class BackendChoiceField(forms.ChoiceField):
    widget = SelectWithSubjectData

    def __init__(self, choices=(), *args, **kwargs):
        backs = list(choices)
        choice_pairs = [(back.ID, back.VERBOSE_NAME) for back in backs]
        use_subjects = dict([(back.ID, back.USE_SUBJECT) for back in backs])
        super().__init__(choices=choice_pairs, *args, **kwargs)
        self.widget.use_subjects = use_subjects


class TemplateInlineFormSet(forms.BaseInlineFormSet):
    def get_form_kwargs(self, index):
        kwargs = super().get_form_kwargs(index)
        kwargs['notification'] = self.instance
        return kwargs


class TemplateForm(forms.ModelForm):
    backend = BackendChoiceField(choices=backends.get_backends())
    recipient = forms.ChoiceField()

    class Meta:
        model = models.Template
        fields = ['notification', 'backend', 'subject', 'content',
                  'recipient', 'is_active']

    def __init__(self, notification=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['recipient'].choices = \
            models.get_recipient_choices(notification)

    def clean(self):
        data = self.cleaned_data
        notification = data['notification']
        try:
            notification.preview(data['backend'], data['content'])
        except Exception as exc:
            raise ValidationError(str(exc))


class SiteContactForm(forms.ModelForm):
    protocol = forms.ChoiceField(choices=backends.get_protocol_choices())

    class Meta:
        model = models.SiteContact
        fields = ['name', 'protocol', 'address']


class NotificationForm(forms.ModelForm):

    class Meta:
        model = models.Notification
        fields = ['codename', 'content_type', 'description', 'target_model']


class NotificationIssueForm(forms.Form):

    contents = forms.ModelMultipleChoiceField(
        queryset=None, label=_('Objects'),
        widget=FilteredSelectMultiple(
            verbose_name=_('Objects'), is_stacked=False),
        help_text=_('A separate notification will be issued for each object.')
    )

    def __init__(self, notification, *args, **kwargs):
        self.notification = notification
        model_cls = notification.content_type.model_class()
        self.declared_fields['contents'].queryset = model_cls.objects
        super().__init__(*args, **kwargs)

    def save(self, commit=True):
        if not commit:
            return
        contents = self.cleaned_data['contents']
        for content in contents:
            self.notification.issue(content)


class TemplateInline(admin.StackedInline):
    model = models.Template
    form = TemplateForm
    formset = TemplateInlineFormSet
    min_num = 0
    extra = 0


class NotificationAdmin(admin.ModelAdmin):
    change_form_template = 'django_courier/notification/change_form.html'
    list_display = ('__str__', 'description', 'required', 'template_count')
    list_filter = ('content_type',)
    inlines = [TemplateInline]
    form = NotificationForm
    issue_form = NotificationIssueForm
    issue_template = 'django_courier/notification/issue.html'

    fields = ['codename', 'content_type', 'description']

    class Media:
        css = {
            'all': ('django_courier/markitup/images.css',
                    'django_courier/markitup/style.css'),
        }
        js = ('django_courier/markitup/jquery.markitup.js',
              'django_courier/notification_fields.js')

    # only show inlines on change forms
    def get_inline_instances(self, request, obj=None):
        return obj and super().get_inline_instances(request, obj) or []

    def get_urls(self):
        return [
            url(
                r'^(?P<id>\w+)/preview/(?P<backend_id>.+)/$',
                self.admin_site.admin_view(self.preview),
                name='django_courier_preview'),
            url(
                r'^(?P<id>\w+)/variables/$',
                self.admin_site.admin_view(self.variables),
                name='django_courier_variables'),
            url(
                r'^(?P<id>\w+)/issue/$',
                self.admin_site.admin_view(self.issue),
                name='django_courier_issue'),
            ] + super().get_urls()

    def preview(self, request, id, backend_id):
        if not self.has_change_permission(request):
            raise PermissionDenied
        notification = self.get_object(request, unquote(id))
        if notification is None:
            raise django.http.Http404(
                _('%(name)s object with primary key %(key)r does not exist.')
                % {'name': force_text(self.model._meta.verbose_name),
                   'key': escape(id)})
        if request.method != 'POST':
            return django.http.HttpResponseNotAllowed(('POST',))
        try:
            result = notification.preview(backend_id, request.POST['body'])
        except Exception as exc:
            result = 'Unable to make preview: {}'.format(str(exc))
        return django.http.HttpResponse(result)

    def variables(self, request, id):
        if not self.has_change_permission(request):
            raise PermissionDenied
        notification = self.get_object(request, unquote(id))
        if notification is None:
            raise django.http.Http404(
                _('%(name)s object with primary key %(key)r does not exist.')
                % {'name': force_text(self.model._meta.verbose_name),
                   'key': escape(id)})
        if request.method != 'POST':
            return django.http.HttpResponseNotAllowed(('POST',))
        result = notification.get_recipient_variables()
        return django.http.JsonResponse(result, safe=False)

    def issue(self, request, id, form_url=''):
        if not self.has_change_permission(request):
            raise PermissionDenied
        notification = self.get_object(request, unquote(id))
        if notification is None:
            raise django.http.Http404(
                _('%(name)s object with primary key %(key)r does not exist.')
                % {'name': force_text(self.model._meta.verbose_name),
                   'key': escape(id)})
        if request.method == 'POST':
            form = self.issue_form(notification, request.POST)
            if form.is_valid():
                form.save()
                msg = ugettext('Notification sent successfully.')
                django.contrib.messages.success(request, msg)
                return django.http.HttpResponseRedirect(
                    reverse(
                        '%s:%s_%s_change' % (
                            self.admin_site.name,
                            notification._meta.app_label,
                            notification._meta.model_name,
                        ),
                        args=(notification.pk,),
                    )
                )
        elif request.method == 'GET':
            form = self.issue_form(notification)
        else:
            return django.http.HttpResponseNotAllowed(('POST',))

        fieldsets = [(None, {'fields': list(form.base_fields)})]
        admin_form = admin.helpers.AdminForm(form, fieldsets, {})

        context = {
            'title': _('Send notification: %s') % escape(str(notification)),
            'media': self.media + admin_form.media,
            'form_url': form_url,
            'form': form,
            'is_popup': (IS_POPUP_VAR in request.POST or
                         IS_POPUP_VAR in request.GET),
            'add': True,
            'change': False,
            'has_delete_permission': False,
            'has_change_permission': True,
            'has_absolute_url': False,
            'opts': self.model._meta,
            'original': notification,
            'save_as': False,
            'show_save': True,
        }
        context.update(self.admin_site.each_context(request))
        request.current_app = self.admin_site.name

        return TemplateResponse(request, self.issue_template, context)

    def get_readonly_fields(self, request, obj=None):
        if self.has_delete_permission(request, obj):
            return ['content_type'] if obj else []
        return ['codename', 'content_type', 'description', 'required',
                'source_model', 'target_model']

    def has_delete_permission(self, request, obj=None):
        if obj is None:
            return super().has_delete_permission(request)
        return super().has_delete_permission(request) and not obj.from_code

    def save_form(self, request, form, change):
        result = super().save_form(request, form, change)
        if not change:
            result.from_code = False
        return result

    @staticmethod
    def template_count(obj):
        return obj.template_set.count()


class SiteContactPreferenceInline(admin.TabularInline):
    model = models.SiteContactPreference


class SiteContactAdmin(admin.ModelAdmin):
    form = SiteContactForm
    inlines = [SiteContactPreferenceInline]
    list_display = ('name', 'address', 'protocol')


class FailedMessageAdmin(admin.ModelAdmin):
    list_display = ('backend', 'address', 'created_at')
    list_filter = ('backend', 'address')


admin.site.register(models.SiteContact, SiteContactAdmin)
admin.site.register(models.FailedMessage, FailedMessageAdmin)
admin.site.register(models.Notification, NotificationAdmin)



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_courier', '0011_notification_from_code'),
    ]

    operations = [
        migrations.AddField(
            model_name='sitecontact',
            name='enable_filter',
            field=models.CharField(
                choices=[('blacklist', 'Blacklist'),
                         ('whitelist', 'Whitelist')],
                default='blacklist', max_length=10),
        ),
    ]



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_courier', '0010_sender_recipient_models'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='from_code',
            field=models.BooleanField(
                default=False, verbose_name='from code',
                help_text='True if the notification is defined in the '
                          'code and automatically managed'),
        ),
        migrations.RemoveField('notification', 'use_sender'),
        migrations.RemoveField('notification', 'sender_model'),
        migrations.RemoveField('notification', 'use_recipient'),
        migrations.RemoveField('notification', 'recipient_model'),
        migrations.AddField(
            model_name='notification',
            name='source_model',
            field=models.ForeignKey(
                blank=True, null=True,
                on_delete=models.deletion.CASCADE,
                related_name='+', to='contenttypes.ContentType',
                verbose_name='source model'),
        ),
        migrations.AddField(
            model_name='notification',
            name='target_model',
            field=models.ForeignKey(
                blank=True, null=True,
                on_delete=models.deletion.CASCADE,
                related_name='+', to='contenttypes.ContentType',
                verbose_name='target model'),
        ),
        migrations.RenameField('template', 'target', 'recipient'),
        migrations.AlterField(
            model_name='template',
            name='recipient',
            field=models.CharField(
                default='re', max_length=103, verbose_name='recipient',
                help_text='Who this message actually gets sent to.'),
        ),
    ]



# Generated by Django 2.0.3 on 2018-03-21 20:10

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_courier', '0005_contact_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='template',
            name='notification',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to='django_courier.Notification',
                verbose_name='notification'),
        ),
    ]



"""
URL patterns for testing app.

We don't have any fore now.
"""
from django.conf.urls import url

from . import views

app_name = 'tests'
urlpatterns = [
    url(r'^$', views.article_list, name='index'),
    url(r'^(?P<article_id>[0-9]+)/$', views.article_detail, name='detail'),
    url(r'^subscribe$', views.subscribe, name='subscribe'),
    url(r'^(?P<article_id>[0-9]+)/comment$', views.comment, name='comment'),
]



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_courier', '0006_django2'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='failedmessage',
            options={'verbose_name': 'failed message'},
        ),
        migrations.AlterModelOptions(
            name='sitecontact',
            options={'verbose_name': 'site contact'},
        ),
        migrations.AlterModelOptions(
            name='template',
            options={'verbose_name': 'template'},
        ),
        migrations.RemoveField(
            model_name='failedmessage',
            name='protocol',
        ),
        migrations.AddField(
            model_name='failedmessage',
            name='contact_name',
            field=models.CharField(
                default='unnamed', max_length=500,
                verbose_name='contact name'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='template',
            name='target',
            field=models.CharField(
                default='re',
                help_text='Who this message actually gets sent to.',
                max_length=103),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.8 on 2017-12-30 07:06
from __future__ import unicode_literals

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import django_courier.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Contact',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True, serialize=False,
                    verbose_name='ID')),
                ('address', models.CharField(max_length=500)),
                ('backend', models.CharField(max_length=100)),
                ('user', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'default_permissions': (),
                'verbose_name': 'contact',
            },
        ),
        migrations.CreateModel(
            name='Notification',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True, serialize=False,
                    verbose_name='ID')),
                ('codename', models.CharField(max_length=100,
                                              verbose_name='codename')),
                ('content_type', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    to='contenttypes.ContentType',
                    verbose_name='content type')),
            ],
            managers=[
                ('objects', django_courier.models.NotificationManager()),
            ],
        ),
        migrations.CreateModel(
            name='Template',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True, serialize=False,
                    verbose_name='ID')),
                ('backend', models.CharField(max_length=100)),
                ('content', models.TextField()),
                ('is_active', models.BooleanField(default=True)),
                ('notification', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    to='django_courier.Notification',
                    verbose_name='notification')),
            ],
            options={
                'default_permissions': (),
                'verbose_name': 'template',
            },
        ),
        migrations.AlterUniqueTogether(
            name='contact',
            unique_together={('user', 'address', 'backend')},
        ),
        migrations.AddField(
            model_name='notification',
            name='use_sender',
            field=models.BooleanField(),
        ),
        migrations.AddField(
            model_name='notification',
            name='use_recipient',
            field=models.BooleanField(default=True),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='notification',
            name='description',
            field=models.TextField(default=''),
            preserve_default=False,
        ),
    ]



import abc
from typing import List


class Contact:
    """A generic contact object

    If you want to return something that looks like this, make sure
    to implement the __hash__ method the same, otherwise filtering
    duplicate contacts won't work
    """

    def __init__(self, name: str, protocol: str, address: str):
        self.name = name
        self.protocol = protocol
        self.address = address

    def __str__(self):
        return '{} <{}:{}>'.format(self.name, self.protocol, self.address)

    def __hash__(self):
        return hash(str(self))


class AbstractContactable(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get_contacts_for_notification(
            self, notification: 'Notification') -> List[Contact]:
        ...



from bs4 import BeautifulSoup
from django.core import mail
from django.test import Client, TestCase

import django_courier.models

from . import models


class DemoTests(TestCase):
    """Test the walkthrough in the documentation"""

    fixtures = ['test']

    @staticmethod
    def test_walkthrough():
        # sanity
        assert len(mail.outbox) == 0
        # first we create an article as the author user
        author = models.User.objects.get(email='author@example.org')
        models.Article.objects.create(
            author=author,
            title='A second article',
            content='Whoever thought we\'d come this far',
        )
        # now a notification should be fired, check the outbox
        # mail.outbox is a list of EmailMultiAlternatives
        assert len(mail.outbox) == 2
        for message in mail.outbox:
            assert (message.content_subtype == 'plain'
                    and message.mixed_subtype == 'mixed')
            html = message.alternatives[0][0]
            soup = BeautifulSoup(html, 'html.parser')
            anchors = soup.find_all('a')
            assert len(anchors) == 1
            url = anchors[0].get('href')
            assert url.startswith('http://127.0.0.1:8000/2/?token=')
            assert len(url) > 31  # if less, token is blank
            client = Client()
            response = client.get(url)
            assert response.status_code == 200
            assert response['Content-Type'] == 'text/html; charset=utf-8'
            # TODO: continue

    @staticmethod
    def test_markdown():
        # sanity
        assert len(mail.outbox) == 0
        # get the article created template
        template = django_courier.models.Template.objects.get(pk=1)
        template.backend = 'email-md'
        template.subject = 'Hi {{ recipient.name }}'
        template.content = 'Hi {{ recipient.name }},\n\nA new article, ' \
                           '[{{ content }}](http://127.0.0.1:8000/{{ ' \
                           'content.pk }}/?token={{ recipient.get_token | ' \
                           'urlencode }}), has been published at the ' \
                           'awesome blog.'
        template.save()
        # first we create an article as the author user
        author = models.User.objects.get(email='author@example.org')
        models.Article.objects.create(
            author=author,
            title='A second article',
            content='Whoever thought we\'d come this far',
        )
        # now a notification should be fired, check the outbox
        # mail.outbox is a list of EmailMultiAlternatives
        assert len(mail.outbox) == 2
        mail_by_subject = {}
        for message in mail.outbox:
            mail_by_subject[message.subject] = message
        site_mail = mail_by_subject['Hi Subscriber']
        assert len(site_mail.alternatives) == 1
        soup = BeautifulSoup(site_mail.alternatives[0][0], 'html.parser')
        anchors = soup.find_all('a')
        assert len(anchors) == 1
        url = anchors[0].get('href')
        assert url.startswith('http://127.0.0.1:8000/2/?token=')
        assert len(url) > 31  # if less, token is blank



from django.contrib.auth.base_user import BaseUserManager
from django.contrib.auth.models import AbstractBaseUser, PermissionsMixin
from django.core import signing
from django.db import models
from django.db.models import Q
from django.utils import crypto
from django.utils.translation import ugettext_lazy as _

from django_courier.base import Contact
from django_courier.models import CourierModel, CourierParam


class UserManager(BaseUserManager):
    def create_user(self, email, name, password=None, **kwargs):
        """
        Creates and saves a User with the given email, date of
        birth and password.
        """
        if not email:
            raise ValueError('Users must have an email address')

        user = self.model(
            email=self.normalize_email(email),
            name=name,
            **kwargs,
        )

        if password is not None:
            user.set_password(password)
        user.save(using=self._db)
        return user

    def create_superuser(self, email, name, password):
        """
        Creates and saves a superuser with the given email, date of
        birth and password.
        """
        user = self.create_user(
            email,
            name=name,
            password=password,
        )
        user.is_superuser = True
        user.is_staff = True
        user.save(using=self._db)
        return user


class User(CourierModel, AbstractBaseUser, PermissionsMixin):

    email = models.EmailField(_('email'), max_length=254, unique=True)
    name = models.CharField(_('name'), max_length=254)
    is_staff = models.BooleanField(
        verbose_name=_('staff status'), default=False,
        help_text=_('Designates whether the user can '
                    'log into this admin site.'))
    is_active = models.BooleanField(
        verbose_name=_('active'), default=True,
        help_text=_('Designates whether this user should be treated as '
                    'active. Un-select this instead of deleting accounts.'))

    objects = UserManager()

    USERNAME_FIELD = 'email'
    REQUIRED_FIELDS = ['name']

    def __str__(self):
        return '{0} <{1}>'.format(self.name, self.email)

    def get_full_name(self):
        return self.name

    def get_short_name(self):
        return self.name

    def get_contacts_for_notification(self, _notification):
        yield Contact(self.name, 'email', self.email)


class Article(CourierModel):

    class CourierMeta:
        notifications = (
            CourierParam(
                'created', _('Notification that a new article was created.'),
            ),
        )

    author = models.ForeignKey(
        to=User, on_delete=models.CASCADE, related_name='+')
    title = models.CharField(_('title'), max_length=254)
    content = models.TextField(_('content'))

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        new = self.id is None
        super().save(*args, **kwargs)
        if new:
            self.issue_notification('created')

    def get_subscribers(self):
        return Subscriber.objects.filter(
            Q(author=self.author) | Q(author=None))

    def get_authors(self):
        yield self.author


class Subscriber(CourierModel):
    """
    A subscriber to blog articles.

    If the author field is set, only articles from those authors
    are subscribed to; otherwise they all are.
    """

    class Meta:
        unique_together = (('author', 'email',),)

    class CourierMeta:
        notifications = (
            CourierParam(
                'created', _(
                    'Notification from subscriber that a new subscriber was '
                    'created. Intended for site admins')),
        )

    author = models.ForeignKey(
        to=User, on_delete=models.CASCADE,
        null=True, blank=True,
        verbose_name=_('author'),)
    name = models.CharField(_('name'), max_length=254)
    email = models.EmailField(_('email'), max_length=254, unique=True)
    secret = models.CharField(
        _('secret'), max_length=32, default='', blank=True,
        help_text=_('Used for resetting passwords'))

    def __str__(self):
        return '{} <{}>'.format(self.name, self.email)

    def save(self, *args, **kwargs):
        new = self.id is None
        super().save(*args, **kwargs)
        if new:
            self.issue_notification('created')

    @classmethod
    def load_from_token(cls, token):
        signer = signing.Signer()
        try:
            unsigned = signer.unsign(token)
        except signing.BadSignature:
            raise ValueError("Bad Signature")

        parts = unsigned.split(' | ')
        if len(parts) < 2:
            raise ValueError("Missing secret or key")
        secret = parts[0]
        email = parts[1]
        user = cls.objects.get(email=email)
        if user.secret != secret:
            raise LookupError("Wrong secret")
        return user

    def get_token(self):
        """Makes a verify to verify new account or reset password

        Value is a signed 'natural key' (email address)
        Reset the token by setting the secret to ''
        """
        if self.secret == '':
            self.secret = crypto.get_random_string(32)
            self.save()
        signer = signing.Signer()
        parts = (self.secret, self.email)
        return signer.sign(' | '.join(parts))

    def get_contacts_for_notification(self, _notification):
        yield Contact(self.name, 'email', self.email)


class Comment(CourierModel):

    class CourierMeta:
        notifications = (
            CourierParam(
                'created', _('Notification that a comment was posted')),
            )

    content = models.TextField(_('content'))
    poster = models.ForeignKey(to=Subscriber, on_delete=models.CASCADE)
    article = models.ForeignKey(to=Article, on_delete=models.CASCADE)

    def save(self, *args, **kwargs):
        new = self.id is None
        super().save(*args, **kwargs)
        if new:
            self.issue_notification('created')

    def get_posters(self):
        yield self.poster

    def get_article_authors(self):
        yield self.article.author


User.add_channel('')
Article.add_channel('sub', _('Subscribers'),
                    Subscriber, Article.get_subscribers)
Article.add_channel('author', _('Author'), User, Article.get_authors)
Subscriber.add_channel('')
Comment.add_channel('poster', _('Poster'), Subscriber, Comment.get_posters)
Comment.add_channel('author', _('Article author'),
                    User, Comment.get_article_authors)
